function [i, A, C1, C2, Na, score, sig, alli, allC1, allNa, allSig] = cvFindLambdaG(D, t, timeVals, classes, inMinCntrlPts, inNa, inMaxCntrlPts)
%[i, A, C1, C2, Na, score, sig, alli, allC1, allNa, allSig] = 
% cvFindLambdaG(D, t, timeVals, classes, inMinCntrlPts, inNa, inMaxCntrlPts)
%
% similar to findLambdaG, which uses GCV. This one uses CV, along with
% a few tricks to speed up things. If the clustering is prespecificed
% (i.e. inNa) is known, then it picks 20 genes from each cluster and
% works with those. Otherwise, it uses maxCntrlPts to identify the
% clustering, and works with that.
%
% inputs:
% D - table containing the observed data (rows = genes, cols = observations)
% t - a vector containing the indices (columns of D) of the experiments we
%     are using as timepoints for the smoothing.  if there are k
%     timepoints represented in D, then t[1] should = 1, 
%     and t[k] should = size(D,2).
% timeVals - the time points (i.e. sampling time) of each
%     experiment represented in D.
% classes - the number of classes to use when smoothing
% inMinCntrlPts - (optional) lower (incl.) limit on number of c.p.s to
% be tried. 0 is always allowed
% inNa - if an input classification is supplied, this is it. zero means ignore
% inMaxCntrlPts: corresponding to inMinCntrlPts, 0 always allowed
% outputs:
% i - the optimal number of control points (i.e., lambda)
% A - the smoothing matrix for the optimal lambda
% C1 - the estimated values (for each gene) at the timepoints t
% C2 - the estimated values (for each gene) at each timepoint in D
% Na - the actual classfication of the genes
% score - the CV score for the optimal lambda
% sig - the variance (not std dev!) in the curves, as calculated by the
% contRep() code
% alli - vector containing all the control points tried
% allC1 - all the C1 matrices corresponding to the different control points tried
% allNa - all the Na matrices corresponding to the different control points tried
% allSig - all the sig (variance!) values corresponding to the different control points tried


global topLevelCost;
global imposeGCVcpLimit;
global numSampleGenesPerClass;

haveClassification = 0;
if (nargin >=6 & sum(inNa)~=0)
  haveClassification = 1;
end

haveMinCP=0;
if (nargin >= 5 & imposeGCVcpLimit == 1)
  haveMinCP = 1;
end

haveMaxCP=0;
if (nargin == 7 & imposeGCVcpLimit == 1)
  haveMaxCP = 1;
end

% make sure we have a row vector
if (size(t, 2) == 1)
  t = t';
end

if (size(timeVals, 2) == 1)
    timeVals = timeVals';
end

% t should [1 ... maxSampleNum]
if (t(1) ~= 1 | t(end) ~= max(t) | t(end) ~= size(D,2) | t(end) ~= length(timeVals))
  error('need t(1) == 1 && t(end) == size(D,2)');
end

% how many timepoints are we using to generate the curve?
n = size(t, 2);

if (n < 5)
  error( 'need size(t) > 5');
end

% minimum number of control points to consider
if (length(t)>5)
  M=5;
else
  M=4;
end
if (haveMinCP)
  minCtrlPts = max([M, inMinCntrlPts]);
else
  minCtrlPts = M;
end

% maximum number of control points to consider
if (haveMaxCP & inMaxCntrlPts>=4)
  maxCtrlPts = min([ inMaxCntrlPts, n-1]);
else
  maxCtrlPts = max([M,n-1]); % n-1 because GCV has troubles at n and
end
%maxCtrlPts = max([4,min([n-1, floor(3*n/4)])]); % n-1 because GCV has troubles at n and
                                % 3*n/4 because ziv's code has troubles
                                % for large n

% pick out the experiment vectors for these timpoints
T = D(:, t);

% pick out the timevalues for these timepoints
times = timeVals(t);

% start with no scores seen
bestScore = realmax;

alli = [];
allC1 = [];
allNa = [];
allSig = [];


% if don't have classification get a classification by using
% maxCntrlPts
prevNa = [];
if (~haveClassification)
  [mC1, mC2, mNa, msig] = contRep(T, D, times, timeVals, maxCtrlPts-2, classes);
  prevNa = mNa;
else
  prevNa = inNa;
end

%for each class pick numSampleGenesPerClass genes,  giving us subT, subD, and subNa
%
if (length(numSampleGenesPerClass)==0)
  numSampleGenesPerClass = 20;
end

%flag = 0
%initialize
y1 = zeros(classes,1);
for j = 1:classes
  s1 = find(prevNa == j);
  k=length(s1);
  y1(j) = min( [ numSampleGenesPerClass k]);
end
subsetSize = sum(y1);


sT = ones( subsetSize, length(times))*100;
sD = ones( subsetSize, length(timeVals))*100;
sNa = zeros(1, subsetSize);

%size(sT)
%size(sD)
%size(sNa)

l=1; h=0; x=0;
for i=1:classes 
 s1 = find(prevNa==i);
 r1 = randperm(length(s1));
 
 if (length(s1) < numSampleGenesPerClass)
   x = length(s1);
 else
   x = numSampleGenesPerClass;
 end
 h = l+x-1;
 
 r2 = r1(1:x);
 s2 = s1(r2);
 sT(l:h,:) = T(s2,:);
 sD(l:h,:) = D(s2,:);
 sNa(l:h) = prevNa(s2);
 l = h+1;
end
%randomize the ordering, just to make sure no locality based effects are introduced  
r3 = randperm(length(sNa));
subT(:,:) = sT(r3,:);
subD(:,:) = sD(r3,:);
subNa = sNa(r3);

%size(sNa)
%size(r3)
%size(sT)
%size(sD)
%size(subT)
%size(subD)

%flag=1
  
%run with these
for i = minCtrlPts:maxCtrlPts
  try
    % do a contWithKnownClasses and get alli allC1 allNa allSig
    [C1, C2, sig] = contWithKnownClasses(T, D, times, timeVals, i-2, classes, prevNa);
    Na = prevNa;
    i1 = i-minCtrlPts + 1;
    alli(i1) = i;
    allC1(i1,:,:) = C1;
    allC2(i1,:,:) = C2;
    allNa(i1,:,:) = Na;
    allSig(i1,:,:) = sig;
%flag=2    
    currScore = 0;
    % for each time point
    for j = 2:(length(times)-1) %can't skip the end entries. hard to extrapolate
      %remove j-th column, do a contWithKnownClasses on the remaining.
      rT(:,:) = [subT(:,1:j-1) , subT(:,j+1:end)];
      rD(:,:) = [subD(:,1:times(j)-1) , subD(:,times(j)+1:end)];
      
      rtimes = [times(1:j-1) times(j+1:end)];
      rtimeVals = [timeVals(1:times(j)-1) timeVals(times(j)+1:end)];
      
%size(rT)
%size(rD)
%size(rtimes)
%size(rtimeVals)
%size(times)
%size(timeVals)
%flag=2.5
      [C1, C2, sig] = contWithKnownClasses(rT, rD, rtimes, rtimeVals, i-2, classes, subNa);
      
      %get estimate for the omitted point. and error w.r.t. to original.add to score
      diffj = subT(:,j) - C2(:,times(j));
%flag=3
      %score it up
      currScore = currScore + (sum(diffj.^2) / length(diffj));
%flag=4
    end
    %compare with best score and update
    if (currScore < bestScore)
      bestScore = currScore;
      bestC1(:,:) = allC1(i1,:,:);
      bestC2(:,:) = allC2(i1,:,:);
      bestNa(:,:) = allNa(i1,:,:);
      besti = i;
    end
  catch
    % grab the error that got us here
    [err, id] = lasterr;

    % make sure the error was caused by trying to use too many ctrl
    % points
    if ((length(strfind(err, 'inter.m:SingularMatrix'))>0) | ...
	(length(strfind(err,'interK.m:SingularMatrix'))>0))    
      % if we encountered an error during the first iteration, then we
      % haven't found *any* lambdas yet, and as such, should fail
      if (i == minCtrlPts)
	error('findLambdaG.m:TooFewTimePoints', 'Try using more timepoints.');
      else
	% otherwise, it's okay to just break at this point and work with
	% the lambdas calculated so far
	fprintf(1,'breaking to next iteration inside findLambdaG\n');
	break;
      end
    else
      fprintf(1,'exiting with error in findLambdaG\n');
      % otherwise, the error was unexpected and we should fail
      error(err);
    end           
  end % end of try..catch
end % end of for loop to iterate over ctrl points

%post process
score = bestScore;
C1 = bestC1;
C2 = bestC2;
i = besti;
S = getS(times, i);
A = S*pinv(S);
Na = bestNa;
