function [i, A, C1, C2, Na, score, sig, alli, allC1, allNa, allSig] = ...
    findLambdaG(D, t, timeVals, classes, inMinCntrlPts, inNa, inMaxCntrlPts)
%[i, A, C1, C2, Na, score, sig, alli, allC1, allNa, allSig] =
%findLambdaG(D, t, timeVals, classes, inMinCntrlPts, inNa)
%
% inputs:
% D - table containing the observed data (rows = genes, cols = observations)
% t - a vector containing the indices (columns of D) of the experiments we
%     are using as timepoints for the smoothing.  if there are k
%     timepoints represented in D, then t[1] should = 1, 
%     and t[k] should = size(D,2).
% timeVals - the time points (i.e. sampling time) of each
%     experiment represented in D.
% classes - the number of classes to use when smoothing
% inMinCntrlPts - (optional) lower (incl.) limit on number of c.p.s to
% be tried. 0 is always allowed
% inNa - if an input classification is supplied, this is it. zero means ignore
% inMaxCntrlPts: corresponding to inMinCntrlPts, 0 always allowed
% outputs:
% i - the optimal number of control points (i.e., lambda)
% A - the smoothing matrix for the optimal lambda
% C1 - the estimated values (for each gene) at the timepoints t
% C2 - the estimated values (for each gene) at each timepoint in D
% Na - the actual classfication of the genes
% score - the GCV score for the optimal lambda
% sig - the variance (not std dev!) in the curves, as calculated by the
% contRep() code
% alli - vector containing all the control points tried
% allC1 - all the C1 matrices corresponding to the different control points tried
% allNa - all the Na matrices corresponding to the different control points tried
% allSig - all the sig (variance!) values corresponding to the different control points tried
global topLevelCost;
global imposeGCVcpLimit;

haveClassification = 0;
if (nargin >=6 & sum(inNa)~=0)
  haveClassification = 1;
end

haveMinCP=0;
if (nargin >= 5 & imposeGCVcpLimit == 1)
  haveMinCP = 1;
end

haveMaxCP=0;
if (nargin == 7 & imposeGCVcpLimit == 1)
  haveMaxCP = 1;
end

% make sure we have a row vector
if (size(t, 2) == 1)
  t = t';
end

if (size(timeVals, 2) == 1)
    timeVals = timeVals';
end

% t should [1 ... maxSampleNum]
if (t(1) ~= 1 | t(end) ~= max(t) | t(end) ~= size(D,2) | t(end) ~= length(timeVals))
  error('need t(1) == 1 && t(end) == size(D,2)');
end

% how many timepoints are we using to generate the curve?
n = size(t, 2);

if (n < 5)
  error( 'need size(t) > 5');
end

% minimum number of control points to consider
if (length(t)>5)
  M=5;
else
  M=4;
end
if (haveMinCP)
  minCtrlPts = max([M, inMinCntrlPts]);
else
  minCtrlPts = M;
end

% maximum number of control points to consider
if (haveMaxCP & inMaxCntrlPts>=4)
  maxCtrlPts = min([ inMaxCntrlPts, n-1]);
else
  maxCtrlPts = max([M,n-1]); % n-1 because GCV has troubles at n and
end
%maxCtrlPts = max([4,min([n-1, floor(3*n/4)])]); % n-1 because GCV has troubles at n and
                                % 3*n/4 because ziv's code has troubles
                                % for large n

% pick out the experiment vectors for these timpoints
T = D(:, t);

% pick out the timevalues for these timepoints
times = timeVals(t);

% start with no scores seen
bestScore = realmax;

alli = [];
allC1 = [];
allNa = [];
allSig = [];
% for each possible number of control points, given the timepoint vector t
for (i = minCtrlPts:maxCtrlPts)
  
  % compute the S matrix for the time points and the
  % number of control points under consideration
  S = getS(times, i);
  
  % compute the convolution matrix A=S*S^+ for this S
  A = S*pinv(S);
  
  % watch for the smoothing code being unhappy about the ratio of ctrl
  % pts to timepoints
  try 
    % compute the continuous representation for these parameters so
    % C1 ends up with the estimated values for each unit interval from
    % times(1) through times(end), whilst C2 gets the estimated values for
    % each unit interval from timeVals(1) through timeVals(end).
    % if we know the clustering already, use it. otherwise find
    % clustering too
    if (haveClassification)
      [C1, C2, sig] = contWithKnownClasses(T, D, times, timeVals, i-2, classes, inNa);
      Na = inNa;
    else
      [C1, C2, Na, sig] = contRep(T, D, times, timeVals, i-2, classes);
    end
    i1 = i-minCtrlPts + 1;
    alli(i1) = i;
    allC1(i1,:,:) = C1;
    allNa(i1,:,:) = Na;
    allSig(i1,:,:) = sig;
    
    % we need to find the estimated values in terms of their minute values.
    % C1 has data for each unit interval from times(1) through times(end), so we
    % need to compute an index into C1 using the units of 'times' as our offset.
    K = C1(:, times - times(1) + 1);

    % compute the GCV score using i control points and size(t) time
    % points
    cost = topLevelCost; %1.0;
    score =  GCV(A, T, K, cost);
    % keep around the info for the best i
    if (score < bestScore)
      bestScore = score;
      bestC1 = C1;
      bestC2 = C2;
      bestNa = Na;
      besti = i;
    end
    
  catch
    % grab the error that got us here
    [err, id] = lasterr;
    
    % make sure the sizes are identical before trying to compare
    %if (size(err, 2) ~= size('inter.m:SingularMatrix', 2))
    % just report the error that was thrown
    %error(err);
    %end
    
    % make sure the error was caused by trying to use too many ctrl
    % points
    if ((length(strfind(err, 'inter.m:SingularMatrix'))>0) | (length(strfind(err,'interK.m:SingularMatrix'))>0))
      % if we encountered an error during the first iteration, then we
      % haven't found *any* lambdas yet, and as such, should fail
      if (i == minCtrlPts)
	error('findLambdaG.m:TooFewTimePoints', 'Try using more timepoints.');
      else
	% otherwise, it's okay to just break at this point and work with
	% the lambdas calculated so far
	fprintf(1,'breaking to next iteration inside findLambdaG\n');
	break;
      end
    else
      fprintf(1,'exiting with error in findLambdaG\n');
      % otherwise, the error was unexpected and we should fail
      error(err);
    end    
    
  end     % end of try ... catch
  
end     % end of for loop to iterate over ctrl points

% return the info for the optimal number of control points, given
% timepoints t
score = bestScore;
C1 = bestC1;
C2 = bestC2;
i = besti;
S = getS(times, i);
A = S*pinv(S);
Na = bestNa;
