function  [lcv_BestCntrlPts, lcv_Scores] = lcv(D, Tindxs, allTimes, ...
					       gcv_besti, gcv_alli, gcv_allC1, gcv_allNa, inMinLcvCntrlPts)
% [lcv_BestCntrlPts, lcv_Scores] = lcv(D, Tindxs, allTimes, gcv_besti, gcv_alli, gcv_allC1, gcv_all) 
%
% run a Local Cross Validation analysis of the given data
% input: 
%
% D: G x N: data matrix . G = num genes, N = number of All sample-able time-points
% Tindxs: 1 x n: the vector describing the indices of sampled times in allTimes and D. 
% allTimes: 1 x N: the vector describing all time-points that are available for
%    sampling. Also includes time-points already sampled
%    The next time point to be sampled should come from one of these
% gcv_besti : the optimal number of control points as found by gcv.
% gcv_alli : contains the # of control points tried by gcv
% gcv_allC1:  gcv_allNa(i) is the func estimate found by by using gcv_alli(i) cntrl points
% gcv_allNa : gcv_allNa(i) is classification found by by using gcv_alli(i) cntrl points
% inMinLcvCntrlPts: inMinLcvCntrlPts(i) is the min number (0 always
% allowed!) of control points
%
% output: n = |Tindxs|,  [G, N] = size(D)
%
% lcv_BestCntrlPts: n x 1. the optimal # of ctrl points, for each
%     time-point in Tindxs, as per LCV
% lcv_Scores: n x 1. the optimal score of LCV, for each time-point in Tindxs

global topLevelCost;
global imposeLCVcpLimit;

if (nargin < 7)
  error('Too few/many input args');
end

haveMinCP = 0;
if (nargin == 8 & imposeLCVcpLimit==1)
  haveMinCP = 1;

end


if (size(Tindxs,1)==1)
  Tindxs = Tindxs';
end
n = size(Tindxs,1);

if (size(allTimes,1)==1)
  allTimes = allTimes';
end
N = size(allTimes,1);

if (Tindxs(1) ~= 1 | Tindxs(end) ~= size(D,2) | length(allTimes) ~= size(D,2))
  error('lcv.m: (Tindxs(1) ~= 1 | Tindxs(end) ~= size(D,2) | length(allTimes) ~= size(D,2))');
  return;
end

[numGenes, numExps] = size(D);

Sgcv = getS( allTimes(Tindxs), gcv_besti);
Agcv = Sgcv * pinv(Sgcv);

lcv_BestCntrlPts = [];
lcv_Scores = [];

j=1;
try
  for i=1:n
    if (haveMinCP)
      minCtrlPts = max([ gcv_alli(1), inMinLcvCntrlPts(i)]);
    else
      minCtrlPts = gcv_alli(1);
    end
    maxCtrlPts = max([ gcv_alli(end) ,n-1]); % n-1 because GCV has troubles at n and

    bestScore = 10e10;
    bestLCVi = 0;
    for r = 1:length(gcv_alli)
      l = gcv_alli(r);  

      %sanity checks
      if (l < gcv_besti-2 | l > gcv_besti+3 |  l < minCtrlPts | l > ...
	  maxCtrlPts)
	continue
      end
	
      Sl = getS( allTimes(Tindxs), l);
      Al = Sl * pinv(Sl);
      
      Al_i = repmat(Al(i,:), numGenes,1);
      
      y = D(:,Tindxs);
      f1(:,:) = gcv_allC1(r,:,allTimes - allTimes(1) + 1);
      fhat = f1(:,Tindxs);
      
      t1 = (y-fhat);
      lcv_score_numrtr = sum(sum(Al_i .* (t1.^2)));
      
      ag_ij = Agcv(i,:);
      al_jj = (diag(Al_i));
      t2 = ag_ij * al_jj;
      cost = topLevelCost;
      lcv_score_denmtr = (1 - cost * t2).^2;
      
      if (lcv_score_denmtr < 1e-6)
	continue;
      end

      lcv_score = lcv_score_numrtr/ lcv_score_denmtr;
      if (lcv_score < bestScore)
	bestScore = lcv_score;
	bestLCVi = l;
      end
    end
    
    if (bestLCVi ~= 0) % no suitable cntrl pt found. go with gcv
      lcv_BestCntrlPts(j) = bestLCVi;
      lcv_Scores(j) = bestScore;
    else
      lcv_BestCntrlPts(j) = gcv_besti;
      lcv_Scores(j) = 10e10;
    end
      j = j+1;
  end
catch
  errmsg = lasterr;
  if(strfind(errmsg, 'inter.m:SingularMatrix'))
    fprintf(2,'Breaking out for i=%d, #cntrl-pts =%d \n', i, l);
  else
    error(errmsg);
  end 
end
