function [nextI, numCntrlPts, lcvCntrlPts, finalCI, C1, Na, sig, lcvalclusterdat] = ...
    pickNext(Din, Tindxs, allTimes, classes, strategy, sigma, prevGcvCntrlPts, prevLcvCntrlPts)
% [nextI, numCntrlPts, C1, Na] = pickNext(D, Tindxs, allTimes, classes, strategy, sigma, prevGcvCntrlPts, prevLcvCntrlPts)
%
% given a dataset and a current set of samples, find the best
% clustering for the current set and the next place to sample
%
% input:
% Din: numGenes x numSampled. the set of observations from currently
%    sampled points
% Tindxs: numSampled x 1. the indices (in allTimes) of the time-points 
%    at which these samples were taken. Tindxs(1) should be 1 and 
%    Tindxs(end) = numTotalSampled
% allTimes: numTotalSamples x 1: the actual time-points at which
%    samples are available
% classes: number of classes to be found
% strategy: an optional string denoting which strategy to use. options are
%   'gcv'  : run GCV and use it to estimate the next point
%   'gcv-cluster: like 
%   'lcv-naive' : use LCV, and to estimate errors at unsampled points,
%   avg error from neighbouring sampled points
%   'lcv-smart' [default] : use LCV, and to estimate errors at
%   unsampled points, estimate error at the point due to neighbouring
%   points (individually) and avg the two errors
% sigma: the variance as predicted [default = 0.2]
% prevGcvCntrlPts: prev number of gcv cntrl pts used (optional, 0 always allowed)
% prevLcvCntrlPts: prev number of lcv cntrl pts used (optional, 0 always allowed)
%
% output:
% nextI: the index of time-point (in allTimes) at which to sample next
% numCntrlPts: for given Tindx, what is the best number of control points
%     to use for clustering
% lcvCntrlPts: 1 x numSampled: the optimal number of control points for
%     each sampled time-point, as determined by lcv
% finalCI: expected error at all possible timepoints in allTimes
% C1: numGenes x numSampled. the estimated expression values for the
%     given samples
% Na: the clustering information
% sig: the actual variance (not std dev!) as estimated for the GCV
global topLevelCost;
global imposeGCVcpLimit;
global imposeLCVcpLimit;

if (nargin < 4)
  error('Insufficient Parameters');
  return;
end

if (nargin < 6)
  sigma = 0.2;
end

if (nargin == 4)
  strategy = 'lcv-smart';
end

haveGCVcp = 0;
if (nargin >=7 & imposeGCVcpLimit==1)
  haveGCVcp = 1;
else
  prevGcvCntrlPts = 0;
end

haveLCVcp = 0;
if (nargin == 8 & imposeLCVcpLimit==1)
  haveLCVcp = 1;
  if (length(prevLcvCntrlPts) ~= length(Tindxs))
    error('need length(prevLcvCntrlPts) ~= length(Tindxs)');
  end
else
  prevLcvCntrlPts = zeros(1, length(Tindxs));
end

if (size(Tindxs,1)==1)
  Tindxs = Tindxs';
end
n = size(Tindxs,1);

if (size(allTimes,1)==1)
  allTimes = allTimes';
end
N = size(allTimes,1);

if (Tindxs(1) ~= 1 |  length(Tindxs) ~= size(Din,2) | Tindxs(end) ~= length(allTimes))
  error('pickNext.m: (Tindxs(1) ~= 1 | length(Tindxs) ~= size(D,2) | Tindxs(end) ~= length(allTimes))');
  return;
end

[numG, expts] = size(Din);

D = rand(numG, N); % tried filling it all with 110. contRep() died. so
                   % filling it with random (fake) values
		   %the subsequent algos like it better if they recieved a	      
		   %full data-matrix from which they can extract elements
D(:,Tindxs) = Din;

if (length(strfind(strategy,'lcv-al-cluster')) == 0)
  [gcvI, Agcv, C1gcv, C2gcv, Na_gcv, gcvScore, sig, gcv_alli, gcv_allC1, ...
   gcv_allNa, gcv_allSig] = findLambdaG(D, Tindxs, allTimes, classes, prevGcvCntrlPts);
  
  fprintf(1,'std dev.: %f\n',sqrt(sig));
  
  numCntrlPts = gcvI;
  
  if (haveGCVcp)
    newGcvCntrlPts = gcvI;
  else
    newGcvCntrlPts = 0;
  end
  C1 = C1gcv(:,allTimes-allTimes(1)+1);
  Na = Na_gcv;
  
  lcvalclusterdat = {};
end

%%%%%%%%%%%%%%%%%%
% FIXIT: both in gcv and lcv-smart, i used n pts for calculating S for
% known sampled points while n+1 points for calculating S for unsampled
% points. is this an issue ??
%%%%%%%%%%%%%%%%%%
initCI = ones(1,N)*-1;  % -1 is pretty much an imposs value for CI
finalCI = initCI;
lcvCntrlPts = [];


if (strfind(strategy,'gcv-cluster'))


  % calculate the overall CIs as suggested by GCV
  for i=1:n
    CIgcv(Tindxs(i)) = CI(Agcv(i,i), sqrt(sig));
  end

  for i=1:n-1
    distance = Tindxs(i+1) - Tindxs(i); % distance to next sampled point
    for k=Tindxs(i)+1:Tindxs(i+1)-1
      Tindxs_k = [Tindxs(1:i) ; k; Tindxs(i+1:end)];
      s_k = getS( allTimes(Tindxs_k), gcvI);
      A_k = s_k * pinv(s_k);
      CIgcv(k) = CI( A_k(i+1,i+1), sqrt(sig));
    end
  end

  finalCI = CIgcv;


  %now we calculate cluster specific CIs
  sumCIs = zeros(1,classes);
  pickIs = zeros(1,classes);

  for j=1:classes
    if (sum(Na_gcv==j) < 1)
      sumCIs(j) = -1;
      x = finalCI;
      x(Tindxs) = -1;
      [maxVal, maxIdx] = max(x);
      pickIs(j) = maxIdx ;
      continue;
    end
    z = sum(Na_gcv==j)
    dclust = D( Na_gcv==j, :);
    size(dclust)
    [cl_gcvI, cl_Agcv, cl_C1gcv, cl_C2gcv, cl_Na_gcv, cl_gcvScore] = findLambdaG(dclust, Tindxs,  allTimes, 1);
    aj = diag(cl_Agcv,0);
    aj1 = aj .^ 0.5;
    sumCIs(j) = sum(aj1);

    ci1 = ones(1,N)*-1;

    for i=1:n
      ci1(Tindxs(i)) = CI(cl_Agcv(i,i), sqrt(sig));
    end

    for i=1:n-1
      distance = Tindxs(i+1) - Tindxs(i); % distance to next sampled point
      for k=Tindxs(i)+1:Tindxs(i+1)-1
        Tindxs_k = [Tindxs(1:i) ; k; Tindxs(i+1:end)];
        s_k = getS( allTimes(Tindxs_k), cl_gcvI);
        A_k = s_k * pinv(s_k);
        ci1(k) = CI( A_k(i+1,i+1), sqrt(sig));
      end
    end

    x = ci1;
    x(Tindxs) = -1;
    [maxVal, maxIdx] = max(x);
    pickIs(j) = maxIdx ;
  end

  [maxVal,maxIdx]= max(sumCIs);
  nextI = pickIs(maxIdx(1));
  return;
end

if (strfind(strategy,'gcv'))

  CIgcv = initCI;
  
  for i=1:n
    CIgcv(Tindxs(i)) = CI(Agcv(i,i), sqrt(sig));
  end
    
  for i=1:n-1
    distance = Tindxs(i+1) - Tindxs(i); % distance to next sampled point
    for k=Tindxs(i)+1:Tindxs(i+1)-1
      Tindxs_k = [Tindxs(1:i) ; k; Tindxs(i+1:end)];
      s_k = getS( allTimes(Tindxs_k), gcvI);
      A_k = s_k * pinv(s_k);
      CIgcv(k) = CI( A_k(i+1,i+1), sqrt(sig));
    end
  end
  
  finalCI = CIgcv;
end

if (strfind(strategy,'lcv-naive'))
    [lcvCntrlPts, lcv_Scores] = lcv(D, Tindxs, allTimes, gcvI, gcv_alli, ...
				    gcv_allC1, gcv_allNa, prevLcvCntrlPts);
    
    CIlcv1 = initCI;
    
    for i=1:n
      s_i = getS( allTimes(Tindxs), lcvCntrlPts(i));
      A_i = s_i * pinv(s_i);
      CIlcv1(Tindxs(i)) = CI(A_i(i,i), sqrt(sig));
    end
    
  for i=1:n-1
    distance = allTimes(Tindxs(i+1)) - allTimes(Tindxs(i)); % distance to next sampled point
    for k=Tindxs(i)+1:Tindxs(i+1)-1
      leftCI = CIlcv1( Tindxs(i) );
      rightCI = CIlcv1( Tindxs(i+1) );
      rightDist = allTimes(Tindxs(i+1)) - allTimes(k);
      leftDist = allTimes(k)-allTimes(Tindxs(i));  
      CIlcv1(k) = ( rightDist*leftCI + leftDist*rightCI ) / distance;
    end
  end

  finalCI = CIlcv1;
end


if (strfind(strategy,'lcv-smart'))
    [lcvCntrlPts, lcv_Scores] = lcv(D, Tindxs, allTimes, gcvI, gcv_alli, ...
				    gcv_allC1, gcv_allNa, prevLcvCntrlPts);
    
    CIlcv2 = initCI;
    
    for i=1:n
      s_i = getS( allTimes(Tindxs), lcvCntrlPts(i));
      A_i = s_i * pinv(s_i);
      CIlcv2(Tindxs(i)) = CI(A_i(i,i), sqrt(sig));
    end

    for i=1:n-1
      distance =  allTimes(Tindxs(i+1)) - allTimes(Tindxs(i)); % distance to next sampled point
      for k=Tindxs(i)+1:Tindxs(i+1)-1
	left_s_i = getS( allTimes( [Tindxs(1:i); k; Tindxs(i+1:end)] ), lcvCntrlPts(i));
	left_a_i = left_s_i * pinv(left_s_i);
	leftCI = CI( left_a_i(i+1,i+1), sqrt(sig));
	
	right_s_i = getS( allTimes( [Tindxs(1:i); k; Tindxs(i+1:end)] ), lcvCntrlPts(i+1));
	right_a_i = right_s_i * pinv(right_s_i);
	rightCI = CI( right_a_i(i+1,i+1), sqrt(sig));

	rightDist = allTimes(Tindxs(i+1)) - allTimes(k);
	leftDist = allTimes(k)-allTimes(Tindxs(i));  
	CIlcv2(k) = ( rightDist*leftCI + leftDist*rightCI ) / distance;
    end
  end

  finalCI = CIlcv2;  
end

x = finalCI;
x(Tindxs) = -1;
[maxVal, maxIdx] = max(x);
nextI = maxIdx;

if (strfind(strategy,'gcv-activelearning'))
  currCIsum = 0;
  for k=1:n-1
    ci1 = CI( Agcv(k,k), sqrt(sig));
    ci2 = CI( Agcv(k+1,k+1), sqrt(sig));
    currCIsum = currCIsum+ (ci1+ci2) * 0.5 * (allTimes(Tindxs(k+1)) - allTimes(Tindxs(k)));
  end
  actLrnScores = ones(1,N) * 10e9; 
  Dj = D;
  for i=1:n-1
    for j=Tindxs(i)+1:Tindxs(i+1)-1
      Dj(:,j) = C1(:,j);
      Tj = [Tindxs(1:i); j; Tindxs(i+1:end)];
      [j_gcv, j_Agcv, j_C1gcv, j_C2gcv, j_Na_gcv, j_gcvScore] = ...
	  findLambdaG(Dj, Tj, allTimes, classes, newGcvCntrlPts, Na);
      jCI =0;
      for k=1: (max(size(Tj)) -1 )
	ci1 = CI( j_Agcv(k,k), sqrt(sig));
	ci2 = CI( j_Agcv(k+1,k+1), sqrt(sig));
	jCI = jCI + (ci1 + ci2) * 0.5 * (allTimes(Tj(k+1)) - allTimes(Tj(k))); %area of trapeziod
      end
      actLrnScores(j) = jCI;
    end
  end
  [minCI, minCIidx] = min(actLrnScores);
  nextI = minCIidx;
  if (minCI > currCIsum)
    fprintf(1,'min(actLrnScores) > current CI: (%f) > (%f)\n', minCI, currCIsum);
  end
  actLrnScores(Tindxs) = currCIsum;
  finalCI = actLrnScores;
  %warning('lcvCntrlPts will be empty');
  %return
end


if (strfind(strategy,'lcv-activelearning'))
   [lcvCntrlPts, lcv_Scores] = lcv(D, Tindxs, allTimes, gcvI, gcv_alli, ...
				   gcv_allC1, gcv_allNa, prevLcvCntrlPts);
   CIlcv3 = finalCI;
   
   for i=1:n
     s_i = getS( allTimes(Tindxs), lcvCntrlPts(i));
     A_i = s_i * pinv(s_i);
     CIlcv3(Tindxs(i)) = CI(A_i(i,i), sqrt(sig));
   end

   currCIsum = 0;
   for k=1:n-1
     ci1 = CIlcv3( Tindxs(k) );
     ci2 = CIlcv3( Tindxs(k+1) );
     currCIsum = currCIsum+ (ci1+ci2) * 0.5 * (allTimes(Tindxs(k+1)) - allTimes(Tindxs(k)));
  end
  actLrnScores = ones(1,N) * 10e9; 
 
  Dj = D;
  for i=1:n-1
    for j=Tindxs(i)+1:Tindxs(i+1)-1
      Dj(:,j) = C1(:,j);
      Tj = [Tindxs(1:i); j; Tindxs(i+1:end)];
      if (haveLCVcp)
	newLcvCntrlPts = [lcvCntrlPts(1:i) 0 lcvCntrlPts(i+1:end)]';
      else
	newLcvCntrlPts = zeros(size(Tj));
      end
      [j_gcv, j_Agcv, j_C1gcv, j_C2gcv, j_Na_gcv, j_gcvScore, j_sig, ...
       j_alli, j_allC1, j_allNa, j_allSig] = findLambdaG(Dj, Tj, allTimes, ...
						  classes, newGcvCntrlPts, Na);
      [lcv4CntrlPts, lcv4_Scores] = lcv(Dj, Tj, allTimes, j_gcv, j_alli, ...
					j_allC1, j_allNa, newLcvCntrlPts);
      CIlcv4 = initCI;
    
      for u=1:(n+1)
	s_u = getS( allTimes(Tj), lcv4CntrlPts(u));
	A_u = s_u * pinv(s_u);
	CIlcv4( Tj(u) ) = CI(A_u(u,u), sqrt(sig));
      end

      cisum2 = 0;
      for u=1:n
	ci1 = CIlcv4( Tj(u) );
	ci2 = CIlcv4( Tj(u+1) );
	cisum2 = cisum2 + (ci1+ci2) * 0.5 * (allTimes(Tj(u+1)) - allTimes(Tj(u)));
      end    
      actLrnScores(j) = cisum2;
    end
  end
  [minCI, minCIidx] = min(actLrnScores);
  nextI = minCIidx;
  if (minCI > currCIsum)
    fprintf(1,'min(actLrnScores) > current CI: (%f) > (%f)\n', minCI, currCIsum);
  end
  actLrnScores(Tindxs) = currCIsum;
  finalCI = actLrnScores;
  %warning('lcvCntrlPts will be empty');
  %return
end



if (strfind(strategy,'lcv-al-cluster'))
%flag=1
allTimes(Tindxs)'
  %do contrep with prevGcvCntrlPts to find clustering
  if (prevGcvCntrlPts < 4)
    prevGcvCntrlPts = 4;
  end
  if (size(allTimes,2)==1)
    allTimes = allTimes';
  end
  if (size(Tindxs,2)==1)
    Tindxs = Tindxs';
  end
  
   [gcvI, Agcv, xC1, xC2, xNa, gcvScore, xsig] = findLambdaG(D, Tindxs, ...
						  allTimes, classes, ...
						  prevGcvCntrlPts, zeros(1,size(D,1)), prevGcvCntrlPts+2);
gcvI   
%size(allTimes), size(allTimes(Tindxs)), size(Din), size(D)
  %[xC1, xC2, xNa, xsig] = contRep( Din, D, allTimes(Tindxs), allTimes, prevGcvCntrlPts-2, classes);
  
  clC1 = xC1;
  clNa = xNa;
%size(clC1), size(clNa)
%flag=2
  % for each cluster
  for c=1:classes
    if (sum(xNa==c) > 1)
      % do a findLambdaG for each cluster, with classes=1
      clear yD yNa;
      yD(:,:) = D( xNa==c, :);
      yNa = ones(1,sum(xNa==c));
size(yD), size(yNa)

      [y_gcv, y_Agcv, y_C1gcv, y_C2gcv, y_Na_gcv, y_gcvScore, y_sig] = ...
	  findLambdaG(yD,Tindxs, allTimes, 1, gcvI,0,gcvI+5);%, yNa);

      clC1( xNa==c, :) = y_C1gcv;
      clNumCntrlPts(c) = y_gcv;
      clLen(c) = sum(xNa==c);
flag=3      
      % score each cluster in terms of error (using gcv error)
      clErr(c)=0;
      n = length(Tindxs);
      CIgcv = initCI;
      for r=1:n
	s_r = getS( allTimes(Tindxs), y_gcv);
	A_r = s_r * pinv(s_r);
	CIgcv(Tindxs(r)) = CI(A_r(r,r), sqrt(y_sig));
      end
  
      currCIsum = 0;
      for r=1:n-1
	ci1 = CIgcv( Tindxs(r) );
	ci2 = CIgcv( Tindxs(r+1) );
	currCIsum = currCIsum+ (ci1+ci2) * 0.5 * (allTimes(Tindxs(r+1)) - allTimes(Tindxs(r)));
      end
      clErr(c) = currCIsum;
%flag=4      
    else
      clLen(c) =sum(xNa==c);
      clNumCntrlPts(c) = 4;
      clErr(c) = 0;
flag=5
    end
  end
flag=5.1
clErr
clNumCntrlPts
  % rank clusters by error
  [sR, iR] = sort(clErr);
  
  % pick a cutoff (??) (say, reject last 4)
  global lcvalcluster_threshold;
  if (length(lcvalcluster_threshold)==0) %undefined ?
    lcvalcluster_threshold =  0.4 * 23;
  end
    
%flag=5.2
  nLower = sum(clErr < lcvalcluster_threshold);
  if (classes <= 10)
    toIgnore = min( [classes-2, nLower] );
  else
    if (nLower >= 3 & nLower <=7)
      toIgnore = nLower;
    else
      toIgnore = classes-5;%5;
    end
  end
  toIgnore = max([toIgnore, 0]);
toIgnore,nLower
%flag=6  
%toIgnore=0
  % subset out the data for the clusters
  classesToUse = classes-toIgnore;
  cCnt =1;
  for c=1:classesToUse
    cidx = iR(end-c+1);
    if (clLen(cidx)>0)
      Dsub(cCnt:cCnt + clLen(cidx) -1,:) = D( xNa==cidx, :);
      Nasub(1,cCnt : cCnt + clLen(cidx) -1) = c;
      cCnt = cCnt + clLen(cidx);
    end
  end
size(Dsub),size(Nasub)
flag=7
  if (imposeGCVcpLimit ==1)
    r1 = iR(toIgnore+1:end);
    n1 = clNumCntrlPts(r1);
    minGCVcp = max([prevGcvCntrlPts , max(n1), gcvI]); %min(n1);
  else
    minGCVcp = 0;
  end
minGCVcp

  %   lcv+al on them...
  [j_gcv, j_Agcv, j_C1gcv, j_C2gcv, j_Na_gcv, j_gcvScore, j_sig, j_alli, j_allC1, j_allNa, j_allSig] = findLambdaG(Dsub, Tindxs, allTimes, classesToUse, minGCVcp, Nasub);
flag=8  
  fprintf(1,'std dev.: %f\n',sqrt(j_sig));

  [j_lcvCntrlPts, j_lcv_Scores] = lcv(Dsub, Tindxs, allTimes, j_gcv, ...
				      j_alli, j_allC1, j_allNa);
j_gcv, j_lcvCntrlPts
flag=9  
  CIlcv = initCI;
  
  n = length(Tindxs);
  for i=1:n
    s_i = getS( allTimes(Tindxs), j_lcvCntrlPts(i));
    A_i = s_i * pinv(s_i);
    CIlcv(Tindxs(i)) = CI(A_i(i,i), sqrt(j_sig));
  end
  
  currCIsum = 0;
  for k=1:n-1
    ci1 = CIlcv( Tindxs(k) );
    ci2 = CIlcv( Tindxs(k+1) );
    currCIsum = currCIsum+ (ci1+ci2) * 0.5 * (allTimes(Tindxs(k+1)) - allTimes(Tindxs(k)));
  end
  actLrnScores = ones(1,N) * 10e9; 
 
  Dj = Dsub;
  for i=1:n-1
    for j=Tindxs(i)+1:Tindxs(i+1)-1
allTimes(j)-allTimes(1)+1
      Dj(:,j) = j_C1gcv(:, allTimes(j)-allTimes(1)+1);
      Tj = [Tindxs(1:i) j Tindxs(i+1:end)];
%size(Dj), size(Tj)
flag=10
      [jj_gcv, jj_Agcv, jj_C1gcv, jj_C2gcv, jj_Na_gcv, jj_gcvScore, jj_sig, ...
       jj_alli, jj_allC1, jj_allNa, jj_allSig] = findLambdaG(Dj, Tj, allTimes, ...
						  classesToUse,j_gcv, ...
						  Nasub,j_gcv+3);
%flag=11
      [lcv4CntrlPts, lcv4_Scores] = lcv(Dj, Tj, allTimes, jj_gcv, jj_alli, ...
					jj_allC1, jj_allNa);
flag=12
      CIlcv4 = initCI;
    
      for u=1:(n+1)
	s_u = getS( allTimes(Tj), lcv4CntrlPts(u));
	A_u = s_u * pinv(s_u);
	CIlcv4( Tj(u) ) = CI(A_u(u,u), sqrt(jj_sig));
      end

      cisum2 = 0;
      for u=1:n
	ci1 = CIlcv4( Tj(u) );
	ci2 = CIlcv4( Tj(u+1) );
	cisum2 = cisum2 + (ci1+ci2) * 0.5 * (allTimes(Tj(u+1)) - allTimes(Tj(u)));
      end    
      actLrnScores(j) = cisum2;
    end
  end
flag=13  
  %   find next timepoint to sample
  [minCI, minCIidx] = min(actLrnScores);
  nextI = minCIidx;
  if (minCI > currCIsum)
    fprintf(1,'min(actLrnScores) > current CI: (%f) > (%f)\n', minCI, currCIsum);
  end
flag=14
  lcvalclusterdat = {Tindxs, allTimes, lcvalcluster_threshold, clC1, clNa, clLen, clErr, ...
		     clNumCntrlPts, toIgnore, nLower, minGCVcp, Dsub, ...
		     Nasub, j_gcv, j_sig, j_C1gcv, j_Na_gcv, j_lcvCntrlPts};
  
  %update data structures
  C1 = clC1(:, allTimes - allTimes(1)+1);
  Na = clNa;
  numCntrlPts = max(clNumCntrlPts);
  finalCI = initCI;
  finalCI(Tindxs) = currCIsum;
  lcvCntrlPts = j_lcvCntrlPts;
  sig = j_sig;
flag=15
end
