function [finalT, finalCI, C1, Na, numCntrlPts] = runSimulation(D, startT, ...
						  allTimes, numFinalSamples, classes, strategy, saveFileName, sigma)
% [finalT, finalD, C1, Na, numCntrlPts] = runSimulation(D, startT,
% allTimes, numFinalSamples, classes, strategy, sigma, saveFileName)
% 
% given a simulated dataset and a strategy, run simulations to find the
% optimal best sampling strategy
% input:
% D: numGenes x numTotalSampled : the entire dataset
% startT: 1 x numInitiallySampled:  the index of starting set of sampled time-points (numInitiallySampled >=5)
% allTimes: all possible sampling locations
% numFinalSamples: the final number of locations to be sampled
% classes: number of categories
% strategy: 'gcv' or 'lcv-smart' or 'lcv-naive'
% saveFileName: the file name to which other auxiliary data should be saved
% sigma: the estimated variance. optional, set to 0.2
% output:
% finalT: 1 x numFinalSamples : the final set of sampled time-points
% finalCI: 1 x numTotalSampled : the C.I.s for the data set
% C1: numGenes x numTotalSampled: estimated function values at each of
% the could-be-sampled time-points.
% Na: numGenes x 1: the classification corresponding to C1
% numCntrlPts : the optimal number of control points used, finally.

if (nargin < 7)
  sigma =  0.2;
end

if (size(startT,1)==1)
  startT = startT';
end

if (size(allTimes,1)==1)
  allTimes = allTimes';
end

[numGenes, numTotalSampled] = size(D);
if (startT(1)~=1 | startT(end) ~= numTotalSampled | size(startT,1) <5 | ...
    numTotalSampled ~= length(allTimes))
  error('runSimulation.m: bad startT or allTimes');
end

initialN = size(startT,1);

if (numFinalSamples < initialN)
  error('runSimulation.m: need numFinalSamples >= size(startT)');
end

global topLevelCost;
global imposeGCVcpLimit;
global imposeLCVcpLimit;
global adaptiveCostStrategy;
global DEBUG;

if (length(topLevelCost)==0)
  topLevelCost=1;
end
if (length(imposeGCVcpLimit)==0)
  imposeGCVcpLimit=1;
end
if (length(imposeLCVcpLimit)==0)
  imposeLCVcpLimit=0;
end
if (length(adaptiveCostStrategy)==0)
  adaptiveCostStrategy=1;
end

% the i-th row of sampledPoints contains a bit-vector describing which
% of i time-points were sampled
sampledPoints= zeros(numFinalSamples,numTotalSampled);
sampledPoints(initialN, startT) = 1;

% gcvCntrlPts(i) indicates optimal # cntrl points when |T|=i
gcvCntrlPts = zeros(numFinalSamples,1);

% gcvC1, gcvNa indicate estimate and classification when |T|=i
gcvC1 = zeros(numFinalSamples, numGenes, numTotalSampled);
gcvNa = zeros(numFinalSamples, numGenes);

% bestC1(i), bestNa(i) indicate estimate and classification which can be
% obtained by using all the numTotalSampled samples but using
% gcvCntrlPts(i) control points
bestC1 = zeros(numFinalSamples, numGenes, numTotalSampled);
bestNa = zeros(numFinalSamples, numGenes);

% ci(i) indicates the confidence intervals at when i time-points have
% been sampled
ci = zeros(numFinalSamples,numTotalSampled);

% lcvCntrlPts(i) contains the optimal number of control points as
% predicted for the i sampled and |T|-i unsampled points 
lcvCntrlPts = zeros(numFinalSamples, numTotalSampled);

realTopLevelCost = topLevelCost;

T = startT;
gcvCP = 0;
lcvCP = zeros(size(T));
lcvalclusterAll = {};
for i=initialN: numFinalSamples
   adaptiveCostStrategy

   if (strfind(strategy,'lcv-al-cluster')) %big time hack to save time
     if (i>=numFinalSamples-1)
       DEBUG=3
     end
   end

   [nextI, xnumCntrlPts, xlcvCntrlPts, xfinalCI, xC1, xNa, sig, lcvalclusterdat] = pickNext( ...
      D(:,T), T, allTimes, classes, strategy, sigma, gcvCP, lcvCP);
  gcvCntrlPts(i) = xnumCntrlPts;
  if (size(xC1,2) ~= length(allTimes))
    error('runSimulation.m: my bad 1');
  end
  gcvC1(i,:,:) =xC1;%(:, allTimes-allTimes(1)+1);
  gcvNa(i,:) = xNa;

  %[yC1,yC2,yNa, sig2] = contRep(D, D, allTimes', allTimes', gcvCntrlPts(i)-2, classes);
  %bestC1(i,:,:) = yC2(:,allTimes-allTimes(1)+1);
  %bestNa(i,:) = yNa;
  
  ci(i,:) = xfinalCI;
  
  if (length(strfind(strategy,'lcv'))>0)
    lcvCntrlPts(i,T) = xlcvCntrlPts;
  end
  lcvalclusterAll{end+1} = lcvalclusterdat;
  
  if (i ~= numFinalSamples)
    
    if (imposeGCVcpLimit)
      gcvCP = xnumCntrlPts;
    else
      gcvCP = 0;
    end
    
    if (imposeLCVcpLimit)
      if (length(strfind(strategy,'lcv'))>0)
	aL = xlcvCntrlPts(T<nextI);
	bL = xlcvCntrlPts(T>nextI);
	lcvCP = [aL 0 bL]';
      end
    else
      lcvCP = zeros(1+length(T), 1);
    end
    
    if (adaptiveCostStrategy)
      if (gcvCntrlPts(i)==gcvCntrlPts(i-1))
	  if (strfind(strategy,'lcv-al-cluster'))
	    topLevelCost = max([ 0.6*realTopLevelCost, 0.8 * topLevelCost])
	  else
	    topLevelCost = max([ 0.7*realTopLevelCost, 0.9 * topLevelCost])
	  end
      else
	topLevelCost = realTopLevelCost
      end
    end
    
    a = T(T<nextI);
    b = T(T>nextI);
    T = [a; nextI; b];
    sampledPoints(i,T) = 1;
  else
    C1= xC1;%(:, allTimes-allTimes(1)+1);
    Na =xNa;
    numCntrlPts = xnumCntrlPts;
    finalT = T;  
    finalCI = xfinalCI;
  end
  interrupted =1;
  save(saveFileName, 'interrupted', 'xnumCntrlPts', 'sampledPoints', ...
       'gcvCntrlPts','gcvNa', 'gcvC1', 'bestNa', 'bestC1', 'ci', 'lcvCntrlPts', 'D', 'numFinalSamples', 'startT', 'classes', 'strategy', 'lcvalclusterAll');
end



save(saveFileName, 'finalT', 'finalCI', 'C1', 'Na', 'numCntrlPts', ...
     'sampledPoints', 'gcvCntrlPts','gcvNa', 'gcvC1', 'bestNa', 'bestC1', 'ci', 'lcvCntrlPts', 'D', 'numFinalSamples', 'startT', 'classes', 'strategy', 'lcvalclusterAll');
